#include <cassert>
#include <iostream>
#include <memory>
#include <string>
#include <chrono>
#include <thread>

#include <curses.h>
#include <pulse/error.h>
#include <pulse/simple.h>

#include "SineSynth.h"

WINDOW *win;

constexpr int SAMPLERATE = 48000;
constexpr size_t BUFSIZE = (48000/1000)*4;

#define SAMPTYPE    uint32_t

WINDOW *
init_curses()
{
    WINDOW *win = initscr();
    nodelay(win, TRUE);
    keypad(win, TRUE);
    cbreak();

    return win;
}

pa_simple *
init_pulse()
{
    static const pa_sample_spec ss = {.format = PA_SAMPLE_S32NE,
                                      .rate = SAMPLERATE,
                                      .channels = 1};

    int error = 0;
    pa_simple *s = pa_simple_new(NULL,
                                 "sinesynth",
                                 PA_STREAM_PLAYBACK,
                                 NULL,
                                 "playback",
                                 &ss,
                                 NULL,
                                 NULL,
                                 &error);
    if (error)
    {
        std::cerr << "Error initializing PulseAudio: " << pa_strerror(error) << " ("
                  << error << ")" << std::endl;
        exit(-1);
    }

    return s;
}

template <typename T>
void smpte_imd_test(SineSynth<T> &synth) {
    synth.clearSynths();
    synth.addSynth(std::make_shared<SineGenerator<T>>(
        SAMPLERATE, 60, 0.8
    ));
    synth.addSynth(std::make_shared<SineGenerator<T>>(
        SAMPLERATE, 7000, 0.2
    ));
}

template <typename T>
void ccif_imd_test(SineSynth<T> &synth, float amplitude) {
    assert(amplitude >= 0 && amplitude <= 1);
    synth.clearSynths();
    synth.addSynth(std::make_shared<SineGenerator<T>>(
        SAMPLERATE, 19000, amplitude / 2
    ));
    synth.addSynth(std::make_shared<SineGenerator<T>>(
        SAMPLERATE, 20000, amplitude / 2
    ));
}

int main(int argc, char *argv[])
{
    SineSynth<SAMPTYPE> synth(SAMPLERATE);
    auto gen = std::make_shared<SineGenerator<SAMPTYPE>>(SAMPLERATE, 1000, 1);
    synth.addSynth(gen);
    SAMPTYPE buf2[BUFSIZE];
    synth.getSamples(BUFSIZE, buf2);
    for (auto i = 0; i < BUFSIZE; i++)
        std::cout << i << "," << (int32_t)buf2[i] << std::endl;

    return 0;

    // auto gen = std::make_shared<SweepGenerator<SAMPTYPE>>(SAMPLERATE, 10, 10000, 5, 1);
    // synth.addSynth(gen);

    // auto gen2 = std::make_shared<JTestGenerator<uint32_t>>(SAMPLERATE, 32);
    // synth.addSynth(gen2);

    //smpte_imd_test(synth);

    auto s = init_pulse();
    auto win = init_curses();

    mvwprintw(win, 0, 0, "Initialized\n");

    int error = 0;
    pa_usec_t latency = pa_simple_get_latency(s, &error);
    wprintw(win,
            "Error: %s (%d) Latency: %d usec\n",
            pa_strerror(error),
            error,
            latency);

    SAMPTYPE buf[BUFSIZE];
    long long i;
    mvwprintw(win, 2, 0, "Loops: ");
    mvwprintw(win, 3, 0, "Freq: ");
    mvwprintw(win, 4, 0, "Ampl: ");
    for (i = 0;; i++)
    {
        auto c = getch();
        switch (c)
        {
        case KEY_UP:
            gen->setFrequency(gen->getFrequency() + 100);
            break;
        case KEY_DOWN:
            gen->setFrequency(gen->getFrequency() - 100);
            break;
        case '+':
            gen->setAmplitude(gen->getAmplitude() + 0.1);
            break;
        case '-':
            gen->setAmplitude(gen->getAmplitude() - 0.1);
            break;
        case 'q':
        case 'Q':
            endwin();
            return 0;
        }

        synth.getSamples(BUFSIZE, buf);
        pa_simple_write(s, buf, sizeof(buf), &error);
        mvwprintw(win, 2, 8, "%6d", i);
        mvwprintw(win, 3, 8, "%6.0f", gen->getFrequency());
        mvwprintw(win, 4, 9, "%1.3f (%f)", gen->getAmplitude(), 20 * log10(gen->getAmplitude()/1));
    }

    endwin();
}